EXPORTS
    aacDecoder_AncDataGet
    aacDecoder_AncDataInit
    aacDecoder_Close
    aacDecoder_ConfigRaw
    aacDecoder_DecodeFrame
    aacDecoder_Fill
    aacDecoder_GetFreeBytes
    aacDecoder_GetLibInfo
    aacDecoder_GetStreamInfo
    aacDecoder_Open
    aacDecoder_SetParam
    aacEncClose
    aacEncEncode
    aacEncGetLibInfo
    aacEncInfo
    aacEncOpen
    aacEncoder_GetParam
    aacEncoder_SetParam
